# 
# Copyright (C) 2003 Carl Johan Lejdfors
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Please send all bugreports to: Carl Johan Lejdfors <calle.lejdfors@cs.lth.se>
# 

from Cg import *
from OpenGL.GL import *
from Numeric import *

import string
import sys
import os
import os.path
import tempfile
import re

class Cg:

    def __init__(self, string):
        self.string = string

    def __call__(self,**args):
        impl = CgImpl(self.string)
        return impl(**args)

    def __getattr__(self, attr):
        return CgImpl(self.string,entry=attr)
     
    def __add__(self, other):
        return Cg(self.string+other.string)

class CgImpl:
    stateParameter = {'modelview': (CG_GL_MODELVIEW_MATRIX,
                                    CG_GL_MATRIX_IDENTITY),
                      'modelviewi': (CG_GL_MODELVIEW_MATRIX,
                                     CG_GL_MATRIX_INVERSE),
                      'modelviewt': (CG_GL_MODELVIEW_MATRIX,
                                     CG_GL_MATRIX_TRANSPOSE),
                      'modelviewit': (CG_GL_MODELVIEW_MATRIX,
                                      CG_GL_MATRIX_INVERSE_TRANSPOSE),
                      'projection': (CG_GL_PROJECTION_MATRIX,
                                     CG_GL_MATRIX_IDENTITY),
                      'projectioni': (CG_GL_PROJECTION_MATRIX,
                                      CG_GL_MATRIX_INVERSE),
                      'projectiont': (CG_GL_PROJECTION_MATRIX,
                                      CG_GL_MATRIX_TRANSPOSE),
                      'projectionit': (CG_GL_PROJECTION_MATRIX,
                                       CG_GL_MATRIX_INVERSE_TRANSPOSE),
                      'texture': (CG_GL_TEXTURE_MATRIX,
                                  CG_GL_MATRIX_IDENTITY),
                      'texturei': (CG_GL_TEXTURE_MATRIX,
                                   CG_GL_MATRIX_INVERSE),
                      'texturet': (CG_GL_TEXTURE_MATRIX,
                                   CG_GL_MATRIX_TRANSPOSE),
                      'textureit': (CG_GL_TEXTURE_MATRIX,
                                    CG_GL_MATRIX_INVERSE_TRANSPOSE),
                      'modelviewproj': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                        CG_GL_MATRIX_IDENTITY),
                      'modelviewproji': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                         CG_GL_MATRIX_INVERSE),
                      'modelviewprojt': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                         CG_GL_MATRIX_TRANSPOSE),
                      'modelviewprojit': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                          CG_GL_MATRIX_INVERSE_TRANSPOSE),
                      'worldviewproj': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                        CG_GL_MATRIX_IDENTITY),
                      'worldviewproji': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                         CG_GL_MATRIX_INVERSE),
                      'worldviewprojt': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                         CG_GL_MATRIX_TRANSPOSE),
                      'worldviewprojit': (CG_GL_MODELVIEW_PROJECTION_MATRIX,
                                          CG_GL_MATRIX_INVERSE_TRANSPOSE),
                      'world': (CG_GL_MODELVIEW_MATRIX,
                                CG_GL_MATRIX_IDENTITY),
                      'worldit': (CG_GL_MODELVIEW_MATRIX,
                                  CG_GL_MATRIX_INVERSE_TRANSPOSE),
                      'viewit': (CG_GL_MODELVIEW_MATRIX,
                                 CG_GL_MATRIX_INVERSE_TRANSPOSE)
                      }
    
    cgContext = cgCreateContext()
    
    def __init__(self, string, entry = 'main'):
        self.string = string
        self.entry = entry

        self.target = None
        self.args = None
        self.kwords = {}

        self.cgTextures = []
        self.cgClientStates = []

    def __call__(self,
                 target = None,
                 args = None,
                 **kwords):
        self.target = target
        self.args = args
        self.kwords = kwords
        return self

    def bindVertex(self, resolver):
        if not self.target:
            self.target = cgGLGetLatestProfile(CG_GL_VERTEX)
        self.bind(resolver)

    def bindFragment(self, resolver):
        if not self.target:
            self.target = cgGLGetLatestProfile(CG_GL_FRAGMENT)
        self.bind(resolver)

    def bind(self, resolver):
        self.compile()

        if not self.prog:
            return 
        cgGLLoadProgram(self.prog)
        cgGLBindProgram(self.prog)
            
        self.setParameters(resolver)

        cgGLEnableProfile(self.target)

    def reset(self):
        cgGLDisableProfile(self.target)
        map(cgGLDisableTextureParameter, self.cgTextures)
        self.cgTextures = []

        map(cgGLDisableClientState, self.cgClientStates)
        self.cgClientStates = []
        
    def compile(self):
        if hasattr(self,'prog'):
            return 

        if self.cgc:
            fd,n = tempfile.mkstemp('.cg')
            f = os.fdopen(fd,'w')
            f.write(self.string)
            f.flush()

            profileDict = { arbvp1 : 'arbvp1', vp20 : 'vp20', \
                            vp30   : 'vp30', arbfp1 : 'arbfp1', \
                            fp20   : 'fp20', fp30   : 'fp30' }
            
            stdin,stdout = os.popen4(self.cgc + \
                                     ' -profile ' + profileDict[self.target] + \
                                     ' -entry ' + self.entry + ' ' + n, 't') 
            errors = stdout.read()

            m = re.findall("(.*)\(([0123456789]+)\) : (.*)", errors)
            
            f.close()
            os.unlink(n)
            
            if m:
                self.prog = None
                raise RuntimeError('Unable to compile program:\n %s' %
                                   string.join(map(lambda (f,ln,e): self.entry + \
                                                   '(%s): %s' % (ln,e), m),'\n'))

        self.prog = cgCreateProgram(self.cgContext,
                                    CG_SOURCE,
                                    self.string,
                                    self.target,
                                    self.entry,
                                    self.args)
        
        if not self.prog:
            raise RuntimeError( 'Unable to create program: %s' % 
                                cgGetErrorString(cgGetError()) )
    
    def setParameters(self, resolver):

        param = cgGetFirstParameter(self.prog, CG_PROGRAM)
        while param:
            name = cgGetParameterName(param)

            ## first use compile-time parameters
            if self.kwords.has_key(name):
                value = self.kwords[name]
                self.setParameter(param,value)

            ## next, look in the resolver
            elif hasattr(resolver, name):
                self.setParameter(param, getattr(resolver, name))

            ## else, check the state matrix parameters
            elif self.stateParameter.has_key(name.lower()):
                matrix, modifier = self.stateParameter[name.lower()]
                cgGLSetStateMatrixParameter(param, matrix, modifier)
                    
            elif cgGetParameterVariability(param) == CG_UNIFORM and \
                 cgGetParameterDirection(param) in [CG_IN, CG_INOUT]:
                ## It is only an error if a uniform parameter cannot be located
                ## Any varying parameters that are not supplied "only" result
                ## in undefined behaviour i.e. it is assumed that the renderer
                ## will provide these variables.
                raise RuntimeError('Unable to find parameter %s!' % name)

            param = cgGetNextParameter(param)

    def setParameter(self, param, value = None):

        cgType = cgGetParameterType(param)

        if (cgGetParameterVariability(param) == CG_VARYING and
            cgType != CG_STRUCT):  
            cgGLEnableClientState(param)
            value = asarray(value, Float)
            cgGLSetParameterPointer(param, value.shape[-1],
                                    GL_DOUBLE, 0, value),
            self.cgClientStates.append(param)
            
        elif cgType == CG_FLOAT:
            cgGLSetParameter1f(param, value)

        elif cgType == CG_FLOAT2:
            cgGLSetParameter2f(param, value[0], value[1])

        elif cgType == CG_FLOAT3:
            cgGLSetParameter3f(param, value[0],
                               value[1], value[2])

        elif cgType == CG_FLOAT4:
            cgGLSetParameter4f(param,
                               value[0], value[1],
                               value[2], value[3])

        elif cgType in [CG_FLOAT1x1, CG_FLOAT1x2, CG_FLOAT1x3,
                        CG_FLOAT1x4, CG_FLOAT2x1, CG_FLOAT2x2,
                        CG_FLOAT2x3, CG_FLOAT2x4, CG_FLOAT3x1,
                        CG_FLOAT3x2, CG_FLOAT3x3, CG_FLOAT3x4,
                        CG_FLOAT4x1, CG_FLOAT4x2, CG_FLOAT4x3,
                        CG_FLOAT4x4]:
            cgGLSetMatrixParameterdc(param, value)

        elif cgType in [CG_SAMPLER1D, CG_SAMPLER2D,
                        CG_SAMPLER3D, CG_SAMPLERCUBE]:
            value.texture.bind()
            cgGLSetTextureParameter(param, value.texture.texNo)
            cgGLEnableTextureParameter(param)
            value.bind()
            self.cgTextures.append(param)

        elif cgType == CG_ARRAY:
            for i in range(cgGetArraySize(param,0)):
                aparam = cgGetArrayParameter(param, i)
                self.setParameter(aparam, value[i])

        elif cgType == CG_STRUCT:
            sparam = cgGetFirstStructParameter(param)
            while sparam:
                name = string.split(cgGetParameterName(sparam),'.')[-1]
                val = None
                if hasattr(value, name):
                    val = getattr(value, name)
                elif hasattr(value, name+'s'):
                    val = getattr(value, name+'s')
                if val:
                    self.setParameter(sparam,val)
                sparam = cgGetNextParameter(sparam)

        else:            
            raise RuntimeError('Unknown type %s', cgGetTypeString(type))



if os.name == 'nt':
    cgcExecutable = 'cgc.exe'
else:
    cgcExecutable = 'cgc'
cgcPath = os.path.join(sys.prefix,'lib','site-packages','Cg',cgcExecutable)

## Test if there is an available cgc in
## the Cg installation path or the system path.
##
## /Calle
try:
    h,n = tempfile.mkstemp('.cg')
    i,o,e = os.popen3(cgcPath + ' -help')

    hasCGC = None
    if o.read():
        hasCGC = cgcPath

    else: ## try the standard path
        i,o,e = os.popen3(cgcExecutable + ' -help')
                
        if o.read():
            hasCGC = cgcExecutable

    CgImpl.cgc = hasCGC

finally:
    os.close(h)
    os.unlink(n)
            




try:
    ## Shader targets
    ## DirectX 9
    vs_2_0 = CG_PROFILE_VS_2_0
    vs_2_x = CG_PROFILE_VS_2_X
    ps_2_0 = CG_PROFILE_PS_2_0
    ps_2_x = CG_PROFILE_PS_2_X
    
    ## DirectX 8
    vs_1_1 = CG_PROFILE_VS_1_1 
    ps_1_1 = CG_PROFILE_VS_1_1
    ps_1_2 = CG_PROFILE_VS_1_2
    ps_1_3 = CG_PROFILE_VS_1_3 

except:
    ## The above profiles are not accessible under Linux
    pass

## OpenGL
vp20 = CG_PROFILE_VP20
fp20 = CG_PROFILE_FP20
vp30 = CG_PROFILE_VP30
fp30 = CG_PROFILE_FP30
arbvp1 = CG_PROFILE_ARBVP1
arbfp1 = CG_PROFILE_ARBFP1

             
