# 
# Copyright (C) 2003 Carl Johan Lejdfors
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Please send all bugreports to: Carl Johan Lejdfors <calle.lejdfors@cs.lth.se>
# 

import FX 
import FX.util
import types

class BasicResolver:

    def __init__(self, nameMap, *contexts):
        self.nameMap = nameMap
        self.contexts = list(contexts)

    def __getattr__(self, attr):
        for ctx in self.contexts:
            if hasattr(ctx, attr):
                return getattr(ctx, attr)
        try:
            attr = self.nameMap[attr]
            for ctx in self.contexts:
                if hasattr(ctx, attr):
                    return getattr(ctx, attr)
        except KeyError:
            pass
        
        raise AttributeError

class Effect:
    """ The top-level effect class. To load an effect from a module
        located in the path use:

          e = Effect('MyEffect')

        This effect can the be used as:

          while e.hasMorePasses():
              ... render geometry ...

        If you subclass this class you need to provide the 'name' and
        'passes' attributes. The moduleName constructor argument
        should not be used as this implies loading the effect from the
        corresponding file. """
        
         
    def __init__(self,
                 moduleName = None,
                 nameMap = {},
                 resolverClass = BasicResolver,
                 **kwords):
        ## update my dictionary and set the name of the effect
        ## if we're using a module-style effect
        if moduleName:
            self.name = moduleName
            self.__dict__.update(__import__(moduleName).__dict__)

            if type(self.technique[0]) == types.ListType:
                self.passes = self.technique[0]
            else:
                self.passes = self.technique
                

        ##allow constructor arguments override of effect parameters
        self.__dict__.update(kwords)

        self._resolverClass = resolverClass
        self._nameMap = nameMap
        self.__passCounter = 0
        
    def setTechnique(self, i):
        self.passes = self.technique[i]
        
    def begin(self, *contexts):
        self._resolver = self._resolverClass(self._nameMap,
                                             self,
                                             *contexts)
        return len(self.passes)

    def doPass(self, i):
        if i > 0:
            FX.State.Reset()
        FX.State.Begin(self._resolver)
        self.passes[i].use(FX.State)
        
    def end(self):
        FX.State.Reset()
        self._resolver = None

    def hasMorePasses(self, *contexts):
        if self.__passCounter == 0:
            self.begin(*contexts)
            
        elif self.__passCounter == len(self.passes):
            self.end()
            self.__passCounter = 0
            return False
        
        self.doPass(self.__passCounter)
        self.__passCounter += 1
        return True

        
class Pass:
    def __init__(self, **kwords):
        self.kwords = kwords
            
    def use(self, state):
        for s,v in self.kwords.items():
            try:

                FX.util.marshalFX(state,s,v)
                       
            except AttributeError, e:
                print "State::ERROR Unknown state '%s': '%s'" % (s,e)

            except AssertionError, e:
                print "An assertion was raised when calling '%s' with '%s': '%s'" % (s,v,e)

            except NotImplementedError, e:
                print "%s not implemented" % key

            except RuntimeError, e:
                print "Setting state '%s' with '%s' failed with: '%s'" % (s,v,e)

            except Exception, e:
                print "Setting state '%s' with '%s' terminated with unknown exception: %s" % (s,v,e)
        
