# 
# Copyright (C) 2003 Carl Johan Lejdfors
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
# 
# Please send all bugreports to: Carl Johan Lejdfors <calle.lejdfors@cs.lth.se>
# 

from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from constants import *
from PIL import *
import PIL.Image
import sys
import os

from FX.util import *

class OpenGLTexture:
    ## This class contain code borrowed from the tex_object.py in the Cg-tree.
    ## /Calle Lejdfors

    basePath = ""
    
    def __init__(self,
                 filename = None,
                 type = TEXTURE_2D ):
        self.filename = filename
        self.type = type
        self.texNo = -1

    def maketex(self, target, mipmap = True):
        w,h = self.size
        data = self.data
            
        if mipmap:
            gluBuild2DMipmaps(target, GL_RGB8, w, h,
                              GL_RGB, GL_UNSIGNED_BYTE, data)
        else:
            glTexImage2D(target, 0, GL_RGB, w, h,
                         0, GL_RGB, GL_UNSIGNED_BYTE, data)

    def open(self, filename):
        return PIL.Image.open(os.path.join(OpenGLTexture.basePath,filename))

    def initialize(self, mipmap = True):
        self.texNo = glGenTextures(1)
        glBindTexture(self.type, self.texNo)

        if self.type == TEXTURE_1D:
            if self.filename:
                image = self.open(self.filename)
                self.size = image.size
                self.data = image.tostring("raw", "RGB", 0, -1)

            data = self.data
            w,h = self.size

            if mipmap:
                gluBuild1DMipmaps(self.type, GL_RGB8, w,
                                  GL_RGB, GL_UNSIGNED_BYTE, data)
            else:
                glTexImage1D(self.type, 0, GL_RGB, w, h,
                             0, GL_RGB, GL_UNSIGNED_BYTE, data)

        elif self.type == TEXTURE_2D:
            if self.filename:
                image = self.open(self.filename)
                self.size = image.size
                self.data = image.tostring("raw", "RGB", 0, -1)
            
            self.maketex(self.type, mipmap)

        elif self.type == TEXTURE_CUBE:
            targets = [ GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
                        GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,
                        GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
                        GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,
                        GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
                        GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB ]
            for file,target in zip(self.filename,targets):
                image = self.open(file)
                self.size = image.size
                self.data = image.tostring("raw", "RGB", 0, -1)
                self.maketex(target,mipmap)
            
        else:
            raise NotImplementedError('Non-1D/2D/CUBE textures are not implemented yet!')

    def textureId(self):
        if self.texNo < 0:
            self.initialize()
        return self.texNo        

    def bind(self):
        texNo = self.textureId()
        glEnable(self.type)
        glBindTexture(self.type, texNo)

    def texParameter(self, pname, i):
        if type(i) == type(int()):
            glTexParameteri(self.type, pname, i);
        elif type(i) == type(float()):
            glTexParameterf(self.type, pname, i);
        elif type(i) == type([]):
            if len(i) > 0:
                if type(i[0]) == type(int()):
                    glTexParameteriv(self.type, pname, i)
                elif type(i[0]) == type(float()):
                    glTexParameterfv(self.type, pname, i)

    def MinFilter(self, val):
        self.texParameter(GL_TEXTURE_MIN_FILTER, val)

    def MagFilter(self, val):
        self.texParameter(GL_TEXTURE_MAG_FILTER, val)

    def WrapS(self, val):
        self.texParameter(GL_TEXTURE_WRAP_S, val)

    def WrapT(self, val):
        self.texParameter(GL_TEXTURE_WRAP_T, val)

    def WrapR(self, val):
        self.texParameter(GL_TEXTURE_WRAP_R, val)

    def TexGenS(self, val):
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, val)
        glEnable(GL_TEXTURE_GEN_S)

    def TexGenT(self, val):
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, val)
        glEnable(GL_TEXTURE_GEN_T)

    def TexGenR(self, val):
        glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, val)
        glEnable(GL_TEXTURE_GEN_R)

    def TextureOp(self, val):
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, val)

    def ColorArg(self, i, val):
        glTexEnvi(GL_TEXTURE_ENV, SOURCE0_RGB_NV+i, GL_TEXTURE)

    def ColorOp(self, val):
        self.TextureOp(val)


class RawOpenGLTexture(OpenGLTexture):

    def __init__(self, array, type = GL_TEXTURE_2D):
        OpenGLTexture.__init__(self,type=type)
        self.data = array.tostring()
        self.size = array.shape

    def initialize(self, mipmap = True):
        self.texNo = glGenTextures(1)
        glBindTexture(self.type, self.texNo)

        if self.type == GL_TEXTURE_1D:
            w,d = self.size
            if mipmap:
                gluBuild1DMipmaps(self.type, GL_RGB8, w,
                                  GL_RGB, GL_FLOAT, self.data)
            else:
                glTexImage1D(self.type, 0, GL_RGB, w, h,
                             0, GL_RGB, GL_FLOAT, self.data)

        elif self.type == GL_TEXTURE_2D:
            w,h,d = self.size
            if d == 3:
                format = GL_RGB
            else:
                format = GL_RGBA
                
            if mipmap:
                gluBuild2DMipmaps(self.type, GL_RGB8, w, h,
                                  format, GL_FLOAT, self.data)
            else:
                glTexImage2D(self.type, format, w, h,
                             0, format, GL_FLOAT, self.data)

        else:
            raise NotImplementedError('Non-1D/2D raw textures are not implemented yet!')


class DynamicOpenGLTexture(OpenGLTexture):
    def __init__(self):
        self.type = TEXTURE_2D
        self.texNo = -1

    def initialize(self):
        self.texNo = glGenTextures(1)
    
    def bind(self):
        OpenGLTexture.bind(self)
        
        width = glutGet(GLUT_WINDOW_WIDTH)
        height = glutGet(GLUT_WINDOW_HEIGHT)

        # Sanitycheck to make sure the window size is power of two in both
        # dimensions.
        n1 = width
        n2 = height

        while (n1 > 1):
            if (n1 % 2):
                break
            n1 >>= 1;

        while (n2 > 1):
            if (n2 % 2):
                break
            n2 >>= 1
            
        if not ((n1 == 1) and (n2 == 1)):
            raise Exception('Window size is not a power of two')

        glCopyTexImage2D(TEXTURE_2D, 0, GL_RGB, 0, 0,
                         width, height, 0)
        
class Sampler:

    def __init__(self,
                 texture,
                 **kwords):
        self.texture = texture
        self.kwords = kwords

    def bind(self):
        self.texture.bind()
        for key,val in self.kwords.items():
            try:
                marshalFX(self.texture, key, val)
            except AttributeError, e:
                print "Unknown texture state: '%s'" % key

        
